"""Download and manipulate COD administrative boundaries."""
import logging
from pathlib import Path

import geopandas as gpd
from fiona.errors import DriverError

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.check_file_existence import check_file_existence
from aatoolbox.utils.hdx_api import load_dataset_from_hdx

logger = logging.getLogger(__name__)


class CodAB(DataSource):
    """
    Work with COD AB (administrative boundaries).

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    """

    def __init__(self, country_config: CountryConfig):
        super().__init__(
            country_config,
            datasource_base_dir="cod_ab",
            is_public=True,
            config_datasource_name="codab",
        )
        self._raw_filepath = (
            self._raw_base_dir / f"{self._country_config.iso3}_"
            f"{self._datasource_base_dir}.shp.zip"
        )

    def download(self, clobber: bool = False) -> Path:
        """
        Download COD AB file from HDX.

        Parameters
        ----------
        clobber : bool, default = False
            If True, overwrites existing COD AB files

        Returns
        -------
        The downloaded filepath

        Examples
        --------
        >>> from aatoolbox import create_country_config, CodAB
        >>> # Download COD administrative boundaries for Nepal
        >>> country_config = create_country_config(iso3="npl")
        >>> codab = CodAB(country_config=country_config)
        >>> npl_cod_shapefile = codab.download()
        """
        return self._download(
            filepath=self._raw_filepath,
            hdx_address=f"cod-ab-{self._country_config.iso3}",
            hdx_dataset_name=self._datasource_config.hdx_dataset_name,
            clobber=clobber,
        )

    def process(self, *args, **kwargs):
        """
        Process COD AB data.

        Method not implemented.
        """
        logger.info("`process()` method not implemented for CodAB.")

    def load(self, admin_level: int) -> gpd.GeoDataFrame:  # type: ignore
        """
        Get the COD AB data by admin level.

        Parameters
        ----------
        admin_level: int
            The administrative level

        Returns
        -------
        COD AB geodataframe with specified admin level

        Raises
        ------
        AttributeError
            If the requested admin level is higher than what is available
        FileNotFoundError
            If the requested filename or layer name are not found

        Examples
        --------
        >>> from aatoolbox import create_country_config, CodAB
        >>>
        >>> # Retrieve admin 2 boundaries for Nepal
        >>> country_config = create_country_config(iso3="npl")
        >>> codab = CodAB(country_config=country_config)
        >>> npl_admin2 = codab.load(admin_level=2)
        """
        admin_level_max = self._datasource_config.admin_level_max
        if admin_level > admin_level_max:
            raise AttributeError(
                f"Admin level {admin_level} requested, but maximum set to "
                f"{admin_level_max} in {self._country_config.iso3.upper()} "
                f"config file"
            )
        return self._load_admin_layer(
            layer_name=self._datasource_config.layer_base_name.format(
                admin_level=admin_level
            )
        )

    def load_custom(self, custom_layer_number: int = 0) -> gpd.GeoDataFrame:
        """
        Get the COD AB data from a custom (non-level) layer.

        Parameters
        ----------
        custom_layer_number: int
            The 0-indexed number of the layer listed in the custom_layer_names
            parameter of the country's config file

        Returns
        -------
        COD AB geodataframe with custom admin level

        Raises
        ------
        AttributeError
            If the requested custom layer number is not available
        FileNotFoundError
            If the requested filename or layer name are not found

        Examples
        --------
        >>> from aatoolbox import create_country_config, CodAB
        >>>
        >>> # Retrieve district boundaries for Nepal
        >>> country_config = create_country_config(iso3="npl")
        >>> codab = CodAB(country_config=country_config)
        >>> npl_district = codab.load_custom(custom_layer_number=0)
        """
        # TODO: possibly merge the two load methods
        try:
            # Ignore mypy for this line because custom_layer_names could be
            # None, but this is handled by the caught exceptions
            layer_name = self._datasource_config.custom_layer_names[
                custom_layer_number
            ]  # type: ignore
        except (IndexError, TypeError) as err:
            raise AttributeError(
                f"{custom_layer_number}th custom layer requested but not "
                f"available in {self._country_config.iso3.upper()} config file"
            ) from err
        return self._load_admin_layer(layer_name=layer_name)

    def _load_admin_layer(self, layer_name: str) -> gpd.GeoDataFrame:
        try:
            return gpd.read_file(f"zip:///{self._raw_filepath / layer_name}")
        except DriverError as err:
            raise FileNotFoundError(
                f"Could not read boundary shapefile. Make sure that "
                f"you have already called the 'download' method and "
                f"that the file {self._raw_filepath} exists. If it does "
                f"exist, please check the validity of the layer name: "
                f"'{layer_name}'."
            ) from err

    @check_file_existence
    def _download(
        self,
        filepath: Path,
        hdx_address: str,
        hdx_dataset_name: str,
        clobber: bool,
    ) -> Path:
        return load_dataset_from_hdx(
            hdx_address=hdx_address,
            hdx_dataset_name=hdx_dataset_name,
            output_filepath=filepath,
        )



"""aa-toolbox: Toolbox for Anticipatory Action.

The ``aatoolbox`` library contains modules designed to assist with
downloading and processing the data required for disaster-related
anticipatory action.
"""
from aatoolbox.config.countryconfig import (
    create_country_config,
    create_custom_country_config,
)
from aatoolbox.datasources.codab.codab import CodAB
from aatoolbox.datasources.fewsnet.fewsnet import FewsNet
from aatoolbox.datasources.iri.iri_seasonal_forecast import (
    IriForecastDominant,
    IriForecastProb,
)
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

from ._version import version as __version__

__all__ = (
    "create_country_config",
    "create_custom_country_config",
    "CodAB",
    "GeoBoundingBox",
    "IriForecastProb",
    "IriForecastDominant",
    "FewsNet",
    "__version__",
)



"""Class to download and load IRI's seasonal forecast."""
import logging
import os
from pathlib import Path

import requests
import xarray as xr
from typing_extensions import Literal

import aatoolbox.utils.raster  # noqa: F401
from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.check_file_existence import check_file_existence
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

logger = logging.getLogger(__name__)

_IRI_AUTH = "IRI_AUTH"


class _IriForecast(DataSource):
    """
    Base class to retrieve IRI's forecast data.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        the bounding coordinates of the area that should be included in the
        data.
    forecast_type: str
        The type of forecast information to download,  can be either "prob" or
        "dominant". If "prob" the data will be retrieved that contains the
        probability per tercile. If "dominant" the data will be retrieved that
        contains only one probability indicating the dominant tercile.
    """

    def __init__(
        self,
        country_config,
        geo_bounding_box: GeoBoundingBox,
        forecast_type: Literal["prob", "dominant"],
    ):
        super().__init__(
            country_config=country_config,
            datasource_base_dir="iri",
            is_public=False,
        )
        # round coordinates to correspond with the grid IRI publishes
        # its data on, which is 1 degree resolution
        # non-rounded coordinates can be given to the URL which then
        # automatically rounds them, but for file saving we prefer to do
        # this ourselves
        self._geobb = geo_bounding_box.round_coords(round_val=1, offset_val=0)
        self._forecast_type = forecast_type

    def download(
        self,
        clobber: bool = False,
    ) -> Path:
        """
        Download the IRI seasonal tercile forecast as NetCDF file.

        To download data from the IRI API, a key is required for
        authentication, and must be set in the ``IRI_AUTH`` environment
        variable. To obtain this key config you need to create an account
        `here.<https://iridl.ldeo.columbia.edu/auth/login>`_.
        Note that this key might be changed over time, and need to be updated
        regularly.

        Parameters
        ----------
        clobber : bool, default = False
            If True, overwrites existing raw files

        Returns
        -------
        The downloaded filepath
        """
        iri_auth = os.getenv(_IRI_AUTH)
        if iri_auth is None:
            raise ValueError(
                f"Environment variable {_IRI_AUTH} is not set and thus cannot "
                f"download the data. Set {_IRI_AUTH} to proceed."
            )
        output_filepath = self._get_raw_path()
        output_filepath.parent.mkdir(parents=True, exist_ok=True)
        url = self._get_url()
        logger.info("Downloading IRI NetCDF file.")
        return self._download(
            filepath=output_filepath,
            url=url,
            iri_auth=iri_auth,
            clobber=clobber,
        )

    def process(self, clobber: bool = False) -> Path:
        """
        Process the IRI forecast.

        Should only be called after the ``download`` method has been
        executed.

        Parameters
        ----------
        clobber : bool, default = False
            If True, overwrites existing processed files

        Returns
        -------
        The processed filepath
        """
        ds = self._load_raw()
        processed_file_path = self._get_processed_path()
        processed_file_path.parent.mkdir(parents=True, exist_ok=True)
        return self._process(
            filepath=processed_file_path, ds=ds, clobber=clobber
        )

    def load(self) -> xr.Dataset:
        """
        Load the IRI forecast data.

        Should only be called after the ``download`` and ``process`` methods
        have been executed.

        Returns
        -------
        The processed IRI dataset
        """
        processed_path = self._get_processed_path()
        try:
            ds = xr.load_dataset(processed_path)
        except FileNotFoundError as err:
            raise FileNotFoundError(
                f"Cannot open the netcdf file {processed_path}. "
                f"Make sure that you have already called the 'process' method "
                f"and that the file {processed_path} exists. "
            ) from err
        # TODO: Save coordinate system to a general config
        return ds.rio.write_crs("EPSG:4326", inplace=True)

    def _get_file_name(self) -> str:
        file_name = (
            f"{self._country_config.iso3}"
            f"_iri_forecast_seasonal_precipitation_tercile_"
            f"{self._forecast_type}_{self._geobb.get_filename_repr(p=0)}.nc"
        )
        return file_name

    def _get_raw_path(self) -> Path:
        return self._raw_base_dir / self._get_file_name()

    def _get_processed_path(self) -> Path:
        return self._processed_base_dir / self._get_file_name()

    def _get_url(self) -> str:

        base_url = (
            "https://iridl.ldeo.columbia.edu/SOURCES/.IRI/.FD/"
            f".NMME_Seasonal_Forecast/.Precipitation_ELR/"
            f".{self._forecast_type}/"
        )
        return (
            f"{base_url}"
            f"X/%28{self._geobb.lon_min}%29%28{self._geobb.lon_max}"
            f"%29RANGEEDGES/"
            f"Y/%28{self._geobb.lat_max}%29%28{self._geobb.lat_min}"
            f"%29RANGEEDGES/"
            "data.nc"
        )

    def _load_raw(self) -> xr.Dataset:
        try:
            ds = xr.load_dataset(
                self._get_raw_path(),
                decode_times=False,
                drop_variables="C",
            )
            return ds
        except ValueError as err:
            raise ValueError(
                f"Cannot open the netcdf file {self._get_raw_path()}. "
                f"This might be due to invalid "
                f"download with wrong authentication. Check the validity of "
                f"the authentication key found in your {_IRI_AUTH} environment"
                f"variable and try to download again. Otherwise make sure the "
                f"correct backend for opening a netCDF file is installed."
            ) from err
        except FileNotFoundError as err:
            raise FileNotFoundError(
                f"Cannot open the netcdf file {self._get_raw_path()}. Make "
                f"sure that you have already called the 'download' method "
                f"and that the file {self._get_raw_path()} exists. "
            ) from err

    @check_file_existence
    def _download(
        self, filepath: Path, url: str, iri_auth: str, clobber: bool
    ) -> Path:

        response = requests.get(
            url,
            # have to authenticate by using a cookie
            cookies={"__dlauth_id": iri_auth},
        )
        with open(filepath, "wb") as out_file:
            out_file.write(response.content)
        return filepath

    @check_file_existence
    def _process(self, filepath: Path, ds, clobber: bool) -> Path:
        # fix dates
        ds.aat.set_time_dim(t_dim="F", inplace=True)
        ds.aat.correct_calendar(inplace=True)
        ds = xr.decode_cf(ds)

        # IRI accepts -180 to 180 longitudes and 0 to 360
        # but automatically converts them to -180 to 180
        # so we don't need to do that
        ds.to_netcdf(filepath)
        return filepath


class IriForecastProb(_IriForecast):
    """
    Class to retrieve IRI's forecast data per tercile.

    The retrieved data contains the probability per
    tercile for the given bounding box.
    Automatically all seasons and leadtimes are
    downloaded.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        the bounding coordinates of the area that should
        be included in the data.

    Examples
    --------
    >>> from aatoolbox import create_country_config, \
    ...     GeoBoundingBox, IriForecastProb
    >>> country_config = create_country_config(iso3="bfa")
    >>> geo_bounding_box = GeoBoundingBox(lat_max=13.0,
    ...                                   lat_min=12.0,
    ...                                   lon_max=-3.0,
    ...                                   lon_min=-2.0)
    >>>
    >>> # Initialize class and retrieve data
    >>> iri = IriForecastProb(country_config,geo_bounding_box)
    >>> iri.download() # Must have IRI_AUTH environment variable set
    >>> iri.process()
    >>>
    >>> iri_data = iri.load()
    """

    def __init__(self, country_config, geo_bounding_box: GeoBoundingBox):
        super().__init__(
            country_config,
            geo_bounding_box=geo_bounding_box,
            forecast_type="prob",
        )


class IriForecastDominant(_IriForecast):
    """
    Class to retrieve IRI's forecast dominant tercile data.

    The retrieved data contains the dominant probability.
    Automatically all seasons and leadtimes are
    downloaded.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
       the bounding coordinates of the area that should
       be included in the data.

    Examples
    --------
    >>> from aatoolbox import create_country_config, \
    ...     GeoBoundingBox, IriForecastDominant
    >>> country_config = create_country_config(iso3="bfa")
    >>> geo_bounding_box = GeoBoundingBox(lat_max=13.0,
    ...                                   lat_min=12.0,
    ...                                   lon_max=-3.0,
    ...                                   lon_min=-2.0)
    >>>
    >>> # Initialize class and retrieve data
    >>> iri = IriForecastDominant(country_config,geo_bounding_box)
    >>> iri.download() # Must have IRI_AUTH environment variable set
    >>> iri.process()
    >>>
    >>> iri_data = iri.load()
    """

    def __init__(
        self, country_config: CountryConfig, geo_bounding_box: GeoBoundingBox
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            forecast_type="dominant",
        )



"""Country configuration setting base class."""
from pathlib import Path
from typing import Dict, Optional, Union

from pydantic import BaseModel, root_validator, validator

from aatoolbox.utils.io import parse_yaml


class CodABConfig(BaseModel):
    """COD AB configuration.

    Parameters
    ----------
    hdx_dataset_name: str
        COD AB dataset name on HDX. Can be found by taking the filename as it
        appears on the dataset page.
    layer_base_name: str
        The base name of the different admin layers, that presumably only
        change by a single custom_layer_number depending on the level. Should
        contain {admin_level} in place of the custom_layer_number.
    admin_level_max: int
        The maximum admin level available in the shapefile.
    custom_layer_names: list, optional
        Any additional layer names that don't fit into the admin level paradigm
    """

    hdx_dataset_name: str
    layer_base_name: str
    admin_level_max: int
    custom_layer_names: Optional[list]

    @validator("layer_base_name")
    def _validate_layer_base_name(cls, layer_base_name):
        if "{admin_level}" not in layer_base_name:
            raise ValueError(
                "The layer base name must contain "
                "an {admin_level} placeholder."
            )
        return layer_base_name


class FewsNetConfig(BaseModel):
    """FEWS NET configuration.

    Parameters
    ----------
    region_name: str
        Name of the region the country belongs to. Needed to download the
        regional FEWS NET data
    """

    # values dictionary gets build in order attributes are listed
    # so first define the dict before the region_name
    region_name_code_mapping: Dict[str, str] = {
        "caribbean-central-america": "LAC",
        "central-asia": "CA",
        "east-africa": "EA",
        "southern-africa": "SA",
        "west-africa": "WA",
    }
    region_name: str

    @validator("region_name")
    def regionname_valid(cls, v, values):
        """Check that regionname is one of the valid ones."""
        valid_regionnames = values["region_name_code_mapping"].keys()
        if v not in valid_regionnames:
            raise ValueError(
                f"Invalid region name: {v}. "
                f"Should be one of {', '.join(valid_regionnames)}"
            )
        return v

    @root_validator(pre=False, skip_on_failure=True)
    def _set_region_code(cls, values) -> dict:
        """Set region code based on region name."""
        values["region_code"] = values["region_name_code_mapping"][
            values["region_name"]
        ]
        return values


class CountryConfig(BaseModel):
    """Country configuration."""

    iso3: str
    codab: Optional[CodABConfig]
    fewsnet: Optional[FewsNetConfig]

    @validator("iso3")
    def _validate_iso3(cls, iso3):
        return _validate_iso3(iso3)


def create_country_config(iso3: str) -> CountryConfig:
    """
    Return a country configuration object from AA Toolbox.

    Parameters
    ----------
    iso3 : str
        Country ISO3, must be exactly 3 letters long

    Returns
    -------
    CountryConfig instance
    """
    iso3 = _validate_iso3(iso3)
    try:
        parameters = parse_yaml(
            Path(__file__).parent.resolve() / f"countries/{iso3}.yaml"
        )
    except FileNotFoundError as err:
        raise FileNotFoundError(
            f"A configuration file for {iso3.upper()} is not yet available "
            f"in AA Toolbox. Try using a custom configuration file with "
            f"create_custom_country_config instead, or contact us to "
            f"request that we add this country."
        ) from err
    return CountryConfig(**parameters)


def create_custom_country_config(filepath: Union[str, Path]) -> CountryConfig:
    """
    Return a custom country configuration object.

    Parameters
    ----------
    filepath: str, pathlib.Path
        Path to the configuration file

    Returns
    -------
    CountryConfig instance
    """
    return CountryConfig(**parse_yaml(filepath))


def _validate_iso3(iso3: str):
    if len(iso3) != 3 or not str.isalpha(iso3):
        raise ValueError("ISO3 must be a three letter string.")
    return iso3.lower()



"""Tests for the DataSource class."""
import pytest

from aatoolbox.datasources.datasource import DataSource


class TestClass(DataSource):
    """Class to inherit from DataSource to enable testing of DataSource."""

    def __init__(
        self,
        country_config,
        datasource_base_dir,
        is_public=False,
        is_global_raw=False,
        is_global_processed=False,
        config_datasource_name=None,
    ):
        super().__init__(
            country_config=country_config,
            datasource_base_dir=datasource_base_dir,
            is_public=is_public,
            is_global_raw=is_global_raw,
            is_global_processed=is_global_processed,
            config_datasource_name=config_datasource_name,
        )

    def download(self):
        """Download method required in DataSource class."""
        raise NotImplementedError

    def process(self):
        """Process method required in DataSource class."""
        raise NotImplementedError

    def load(self):
        """Load method required in DataSource class."""
        raise NotImplementedError


def test_config_attribute_name_validator(mock_country_config):
    """Test that correctly checked that datasource_name in the config."""
    with pytest.raises(AttributeError):
        TestClass(
            country_config=mock_country_config,
            datasource_base_dir="fake_dir_name",
            config_datasource_name="fake_config_attribute",
        )

    testclass_valid_attribute_name = TestClass(
        country_config=mock_country_config,
        datasource_base_dir="fewsnet",
        config_datasource_name="fewsnet",
    )
    assert (
        testclass_valid_attribute_name._datasource_config.region_name
        == "east-africa"
    )



"""Tests for the country configuration."""
import copy
from typing import Dict, List

import pytest

from aatoolbox import create_country_config
from aatoolbox.config.countryconfig import FewsNetConfig


@pytest.fixture
def mock_parse_yaml(mocker):
    """Fixture to return custom dict from yaml parser."""

    def patcher(output_list: List[Dict] = None):
        if output_list is None:
            output_list = [{"iso3": "abc"}]
        return mocker.patch(
            "aatoolbox.config.countryconfig.parse_yaml",
            side_effect=output_list,
        )

    return patcher


def test_create_country_config(mock_parse_yaml):
    """Test that country config is created as expected."""
    mock_parse_yaml()
    country_config = create_country_config(iso3="abc")
    assert country_config.iso3 == "abc"


def test_country_not_implemented():
    """Test error raised when country not implemented."""
    with pytest.raises(FileNotFoundError):
        create_country_config(iso3="abc")


@pytest.fixture
def bad_iso3s():
    """Input ISO3s that should throw an error."""
    return ["abcd", "ab", "123"]


def test_input_iso3(bad_iso3s):
    """Test that user input iso3 is validated, but can be uppercase."""
    with pytest.raises(ValueError):
        for iso3 in bad_iso3s:
            create_country_config(iso3=iso3)


def test_uppercase_input_iso3(mock_parse_yaml):
    """Test that uppercase iso3 converted to lowercase for filename."""
    fake_parser = mock_parse_yaml()
    create_country_config(iso3="ABC")
    _, args, _ = fake_parser.mock_calls[0]
    assert "abc.yaml" in str(args[0])


def test_config_iso3(mock_parse_yaml, bad_iso3s):
    """Test that iso3 in config file is validated."""
    mock_parse_yaml(output_list=[{"iso3": iso3} for iso3 in bad_iso3s])
    with pytest.raises(ValueError):
        for _ in bad_iso3s:
            create_country_config(iso3="abc")


def test_uppercase_config_iso3(mock_parse_yaml):
    """Test that uppercase iso3 in config file is converted to lowercase."""
    mock_parse_yaml(output_list=[{"iso3": "ABC"}])
    country_config = create_country_config(iso3="abc")
    assert country_config.iso3 == "abc"


def test_validate_codab_layer_base_name(mock_parse_yaml):
    """Test that layer basename requires correct placeholder."""
    config_base = {
        "iso3": "abc",
        "codab": {
            "hdx_dataset_name": "fake_dataset_name",
            "layer_base_name": "layer_base_name_",
            "admin_level_max": 1,
        },
    }
    config_correct = copy.deepcopy(config_base)
    config_correct["codab"]["layer_base_name"] += "{admin_level}"
    config_incorrect = copy.deepcopy(config_base)
    config_incorrect["codab"]["layer_base_name"] += "{wrong_placeholder}"
    mock_parse_yaml(output_list=[config_correct, config_incorrect])
    # Check that correct config runs without issue
    create_country_config(iso3="abc")
    # Check that incorrect config raises error
    with pytest.raises(ValueError):
        create_country_config(iso3="abc")


def test_validate_fewsnet_region_name(mock_parse_yaml):
    """Test that fewsnet requires correct region name."""
    config_base = {
        "iso3": "abc",
        "fewsnet": {"region_name": ""},
    }
    config_correct = copy.deepcopy(config_base)
    config_correct["fewsnet"]["region_name"] += "east-africa"
    config_incorrect = copy.deepcopy(config_base)
    config_incorrect["fewsnet"]["region_name"] += "fake_region_name"
    mock_parse_yaml(output_list=[config_correct, config_incorrect])
    # Check that correct config runs without issue
    country_config_valid = create_country_config(iso3="abc")
    assert country_config_valid.fewsnet == FewsNetConfig(
        region_name_code_mapping={
            "caribbean-central-america": "LAC",
            "central-asia": "CA",
            "east-africa": "EA",
            "southern-africa": "SA",
            "west-africa": "WA",
        },
        region_name="east-africa",
        region_code="EA",
    )
    # Check that config with incorrect region_name raises error
    with pytest.raises(ValueError):
        create_country_config(iso3="abc")



"""Tests for check_file_existence decorator."""

import pytest

from aatoolbox.utils.check_file_existence import check_file_existence


@check_file_existence
def downloader(filepath, clobber):
    """I/O function for testing."""
    return "a"


def test_fp_exists_no_clobber(tmp_path):
    """Test that filepath returned if Path exists and clobber False."""
    tmp_path.touch()
    output_filepath = downloader(filepath=tmp_path, clobber=False)
    assert output_filepath == tmp_path


def test_fp_exists_clobber(tmp_path):
    """Test that function returned if Path exists and clobber True."""
    tmp_path.touch()
    output_filepath = downloader(filepath=tmp_path, clobber=True)
    assert output_filepath == "a"


def test_fp_not_exists(tmp_path):
    """Test that function returned if Path does not exist."""
    output_filepath = downloader(filepath=tmp_path, clobber=True)
    assert output_filepath == "a"


def test_key_error(tmp_path):
    """Test that KeyError raised when filepath or clobber not a kwarg."""
    with pytest.raises(KeyError):
        downloader(filepath=tmp_path)
        downloader(clobber=True)
        downloader()



"""Tests for the IRI module."""

from pathlib import Path

import cftime
import numpy as np
import pytest
import xarray as xr
from xarray.coding.cftimeindex import CFTimeIndex

from aatoolbox import GeoBoundingBox, IriForecastDominant, IriForecastProb

DATASOURCE_BASE_DIR = "iri"
FAKE_IRI_AUTH = "FAKE_IRI_AUTH"


@pytest.fixture
def mock_iri(mock_country_config):
    """Create IRI class with mock country config."""
    geo_bounding_box = GeoBoundingBox(
        lat_max=6, lat_min=3.2, lon_max=2, lon_min=-3
    )

    def _mock_iri(prob_forecast: bool = True):
        if prob_forecast:
            iri = IriForecastProb(
                country_config=mock_country_config,
                geo_bounding_box=geo_bounding_box,
            )
        else:
            iri = IriForecastDominant(
                country_config=mock_country_config,
                geo_bounding_box=geo_bounding_box,
            )
        return iri

    return _mock_iri


@pytest.fixture
def mock_download(mocker, mock_iri):
    """
    Call download with mocked _download.

    `forecast_type` is the type of forecast to
    test, can be either 'prob' or 'dominant'.
    """
    download_mock = mocker.patch(
        (
            "aatoolbox.datasources.iri.iri_seasonal_forecast"
            "._IriForecast._download"
        )
    )

    mocker.patch.dict(
        "aatoolbox.datasources.iri.iri_seasonal_forecast.os.environ",
        {"IRI_AUTH": FAKE_IRI_AUTH},
    )

    def _mock_download(prob_forecast: bool):
        iri = mock_iri(prob_forecast=prob_forecast)
        iri.download()
        _, kwargs_download = download_mock.call_args
        url = kwargs_download["url"]
        filepath = kwargs_download["filepath"]
        return url, filepath

    return _mock_download


def test_download_call_prob(
    mock_download, mock_aa_data_dir, mock_country_config
):
    """Test download for tercile probability forecast."""
    url, filepath = mock_download(prob_forecast=True)
    assert url == (
        "https://iridl.ldeo.columbia.edu/SOURCES/.IRI/.FD/"
        ".NMME_Seasonal_Forecast/"
        ".Precipitation_ELR/.prob/X/%28-3.0%29%282.0%29RANGEEDGES/"
        "Y/%286.0%29%283.0%29RANGEEDGES/data.nc"
    )

    assert filepath == (
        mock_aa_data_dir
        / f"private/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
        f"abc_iri_forecast_seasonal_precipitation_"
        f"tercile_prob_Np6Sp3Ep2Wm3.nc"
    )


def test_download_call_dominant(
    mock_download, mock_aa_data_dir, mock_country_config
):
    """Test download for dominant tercile forecast."""
    url, filepath = mock_download(prob_forecast=False)
    assert url == (
        "https://iridl.ldeo.columbia.edu/SOURCES/.IRI/.FD/"
        ".NMME_Seasonal_Forecast/"
        ".Precipitation_ELR/.dominant/X/%28-3.0%29%282.0%29RANGEEDGES/"
        "Y/%286.0%29%283.0%29RANGEEDGES/data.nc"
    )

    assert filepath == (
        mock_aa_data_dir
        / f"private/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
        f"abc_iri_forecast_seasonal_"
        f"precipitation_tercile_dominant_Np6Sp3Ep2Wm3.nc"
    )


def test_process(mocker, mock_iri, mock_aa_data_dir, mock_country_config):
    """Test process for IRI forecast."""
    ds = xr.DataArray(
        np.reshape(a=np.arange(16), newshape=(2, 2, 2, 2)),
        dims=("L", "X", "Y", "F"),
        coords={
            "L": [1, 2],
            "X": [2, -3],
            "Y": [97, 90],
            "F": [685.5, 686.5],
        },
    ).to_dataset(name="prob")

    ds["F"].attrs["calendar"] = "360"
    ds["F"].attrs["units"] = "months since 1960-01-01"

    iri = mock_iri()

    # TODO: now created `load_raw` to be able to mock but would like
    #  to do it from xr.load_dataset directly
    mocker.patch(
        "aatoolbox.datasources.iri.iri_seasonal_forecast."
        "_IriForecast._load_raw",
        return_value=ds,
    )

    processed_path = iri.process()
    assert processed_path == (
        mock_aa_data_dir / f"private/processed/{mock_country_config.iso3}/"
        f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}_"
        f"iri_forecast_seasonal_precipitation_"
        f"tercile_prob_Np6Sp3Ep2Wm3.nc"
    )

    da_processed = xr.load_dataset(processed_path)
    expected_f = CFTimeIndex(
        [
            cftime.datetime(year=2017, month=2, day=16, calendar="360_day"),
            cftime.datetime(year=2017, month=3, day=16, calendar="360_day"),
        ]
    )

    assert np.array_equal(da_processed.X.values, [2, -3])
    assert np.array_equal(da_processed.Y.values, [97, 90])
    assert da_processed.get_index("F").equals(expected_f)
    assert np.array_equal(da_processed.prob.values, ds.prob.values)


def test_process_if_download_not_called(mock_iri):
    """Test that correct error message raised."""
    iri = mock_iri()
    # Make sure file doesn't exist
    if iri._get_raw_path().exists():
        Path.unlink(iri._get_raw_path())
    with pytest.raises(FileNotFoundError) as excinfo:
        iri.process()
    assert (
        "Make sure that you have already called the 'download' method"
        in str(excinfo.value)
    )


@pytest.fixture
def mock_xr_load_dataset(mocker):
    """Mock GeoPandas file reading function."""
    return mocker.patch(
        "aatoolbox.datasources.iri.iri_seasonal_forecast.xr.load_dataset"
    )


def test_iri_load(
    mocker,
    mock_xr_load_dataset,
    mock_iri,
    mock_aa_data_dir,
    mock_country_config,
):
    """Test that load_codab calls the HDX API to download."""
    mocker.patch(
        (
            "aatoolbox.datasources.iri.iri_seasonal_forecast"
            "._IriForecast._download"
        )
    )

    iri = mock_iri()
    iri.load()
    mock_xr_load_dataset.assert_has_calls(
        [
            mocker.call(
                mock_aa_data_dir
                / f"private/processed/{mock_country_config.iso3}/"
                f"{DATASOURCE_BASE_DIR}/{mock_country_config.iso3}"
                f"_iri_forecast_seasonal_precipitation_"
                f"tercile_prob_Np6Sp3Ep2Wm3.nc"
            ),
        ]
    )


def test_load_if_process_not_called(mock_iri):
    """Test that correct error message raised."""
    iri = mock_iri()
    # Make sure file doesn't exist
    if iri._get_processed_path().exists():
        Path.unlink(iri._get_processed_path())
    with pytest.raises(FileNotFoundError) as excinfo:
        iri.load()
    assert (
        "Make sure that you have already called the 'process' method"
        in str(excinfo.value)
    )



"""Path configuration."""

import os
from dataclasses import dataclass
from pathlib import Path

_BASE_DIR_ENV = "AA_DATA_DIR"


@dataclass
class PathConfig:
    """Global directory parameters."""

    base_dir_env: str = _BASE_DIR_ENV
    public: str = "public"
    private: str = "private"
    raw: str = "raw"
    processed: str = "processed"

    def __post_init__(self):
        """Set the base path based on the environemnt variable."""
        self.base_path = Path(os.environ[self.base_dir_env])



"""
Utilities to manipulate and analyze raster data.

This module extends xarray and rioxarray to provide
additional functionality for raster processing and
post-processing. The extension is based on the
guidance for how to `extend xarray
<http://xarray.pydata.org/en/stable/internals/extending-xarray.html>`_:

However, since rioxarray already extends xarray, this
modules extensions inherit from the RasterArray and
RasterDataset extensions respectively. This ensures
cleaner code in the module as ``rio`` methods are
available immediately, but also means a couple of
design decisions are followed.

The xarray.DataArray and xarray.Dataset
extensions here inherit from rioxarray base classes.
Thus, methods that are identical for both objects
are defined in a mixin class ``AatRasterMixin`` which
can be inherited by the two respective extensions.
"""

import logging
from typing import Any, Callable, Dict, List, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import rioxarray  # noqa: F401
import xarray as xr
from rioxarray.exceptions import DimensionError, MissingCRS, NoDataInBounds
from rioxarray.raster_array import RasterArray
from rioxarray.raster_dataset import RasterDataset
from rioxarray.rioxarray import _get_data_var_message

logger = logging.getLogger(__name__)


class AatRasterMixin:
    """AA toolbox mixin base class."""

    # setting attributes to avoid mypy error, from SO
    # https://stackoverflow.com/questions/53120262/mypy-how-to-
    # ignore-missing-attribute-errors-in-mixins/53228204#53228204
    _get_obj: Callable
    x_dim: str
    y_dim: str

    def __init__(self, xarray_obj):
        super().__init__(xarray_obj)

        # Adding lat/lon to set of default spatial dims
        if "lat" in self._obj.dims and "lon" in self._obj.dims:
            self._x_dim = "lon"
            self._y_dim = "lat"

        # Adding Y/X to set of default spatial dims
        if "Y" in self._obj.dims and "X" in self._obj.dims:
            self._x_dim = "X"
            self._y_dim = "Y"

        # Managing time coordinate default dims
        self._t_dim = None
        for t in ["t", "T", "time"]:
            if t in self._obj.dims:
                self._t_dim = t

        # longitude range indicates if coordinates are
        # between -180 and 180 or 0 and 360
        self._longitude_range = None

    # methods derived from rioxarray.rioxarray.x_dim and y_dim
    @property
    def t_dim(self):
        """str: The dimension for time."""
        if self._t_dim is None:
            raise DimensionError(
                "Time dimension not found. 'aat.set_time_dim()' or "
                "using 'rename()' to change the dimension name to "
                f"'t' can address this.{_get_data_var_message(self._obj)}"
            )
        return self._t_dim

    @property
    def longitude_range(self):
        """str: The longitude range.

        The longitude range indicates if coordinates are
        between -180 and 180 (indicated by '180')
        or 0 and 360 (indicated by '360').
        """
        if self._longitude_range is None:
            if self._x_dim is not None:
                data_obj = self._get_obj(inplace=False)
                lon_max = data_obj.indexes[self.x_dim].max()
                if lon_max > 180:
                    self._longitude_range = "360"
                else:
                    self._longitude_range = "180"
            else:
                raise DimensionError(
                    "Longitude range not set due to missing 'x_dim'."
                    "Use 'aat.set_spatial_dims()' to set the name of "
                    f"the x dimension. {_get_data_var_message(self._obj)}"
                )
        return self._longitude_range

    def set_time_dim(
        self, t_dim: str, inplace: bool = False
    ) -> Optional[Union[xr.DataArray, xr.Dataset]]:
        """Set the time dimension of the dataset.

        Parameters
        ----------
        t_dim: str
            The name of the time dimension.
        inplace: bool, optional
            If True, it will modify the dataarray in place.
            Otherwise it will return a modified copy.

        Returns
        -------
        Union[xarray.DataArray, xarray.Dataset]
            Data array or dataset with time dimension.

        Examples
        --------
        >>> import xarray
        >>> import numpy
        >>> da = xarray.DataArray(
        ...  numpy.arange(64).reshape(4,4,4),
        ...  coords={"lat":numpy.array([87, 88, 89, 90]),
        ...          "lon":numpy.array([5, 120, 199, 360]),
        ...          "F":numpy.array([10,11,12,13])}
        ... )
        >>> da.aat.set_time_dim(t_dim="F", inplace=True)
        >>> da.aat.t_dim
        'F'
        """
        data_obj = self._get_obj(inplace=inplace)
        if t_dim not in data_obj.dims:
            raise DimensionError(
                "Time dimension ({t_dim}) not found."
                f"{_get_data_var_message(data_obj)}"
            )
        data_obj.aat._t_dim = t_dim
        return data_obj if not inplace else None

    def correct_calendar(
        self, inplace: bool = False
    ) -> Optional[Union[xr.DataArray, xr.Dataset]]:
        """Correct calendar attribute for recognition by xarray.

        Some datasets come with a wrong calendar attribute that isn't
        recognized by xarray. This function corrects the coordinate
        attribute to ensure that a ``calendar`` attribute exists
        and specifies a calendar alias that is supportable by
        ``xarray.cftime_range`` and NetCDF in general.

        Currently ensures that calendar attributes that are either
        specified with ``units="months since"`` or ``calendar="360"``
        explicitly have ``calendar="360_day"``. This is based on
        discussions in `this GitHub issue
        <https://github.com/Unidata/netcdf4-python/issues/434>`_.
        If and when further issues are found with calendar
        attributes, support for conversion will be added here.

        Parameters
        ----------
        inplace: bool, optional
            If True, it will modify the dataarray in place.
            Otherwise it will return a modified copy.

        Returns
        -------
        Union[xarray.DataArray, xarray.Dataset]
            Data array or dataset with transformed calendar
            coordinate.

        Examples
        --------
        >>> import xarray
        >>> import numpy
        >>> da = xarray.DataArray(
        ...  numpy.arange(64).reshape(4,4,4),
        ...  coords={"lat":numpy.array([87, 88, 89, 90]),
        ...          "lon":numpy.array([5, 120, 199, 360]),
        ...          "t":numpy.array([10,11,12,13])}
        ... )
        >>> da["t"].attrs["units"] = "months since 1960-01-01"
        >>> da_crct = da.aat.correct_calendar()
        >>> da_crct["t"].attrs["calendar"]
        '360_day'
        """
        data_obj = self._get_obj(inplace=inplace)
        if (
            "calendar" in data_obj[self.t_dim].attrs.keys()
            and data_obj[self.t_dim].attrs["calendar"] == "360"
        ):
            data_obj[self.t_dim].attrs["calendar"] = "360_day"
            logger.info("Calendar attribute changed from '360' to '360_day'.")

        elif (
            "units" in data_obj[self.t_dim].attrs.keys()
            and "months since" in data_obj[self.t_dim].attrs["units"]
        ):
            data_obj[self.t_dim].attrs["calendar"] = "360_day"
            logger.info(
                "Calendar attribute '360_day' added, "
                "equivalent of 'units' 'months since'."
            )

        else:
            logger.info("No 'units' or 'calendar' attributes to correct.")

        return data_obj if not inplace else None

    def invert_coordinates(self, inplace: bool = False) -> xr.DataArray:
        """
        Invert latitude and longitude in data array.

        This function checks for inversion of latitude and longitude
        and inverts them if needed. Datasets with inverted coordinates
        can produce incorrect results in certain functions like
        ``rasterstats.zonal_stats()``. Correctly ordered coordinates
        should be:

        * latitude: Largest to smallest.
        * longitude: Smallest to largest.

        If data array already has correct coordinate ordering, it is
        directly returned. Function largely copied from
        https://github.com/perrygeo/python-rasterstats/issues/218.

        Parameters
        ----------
        inplace : bool, optional
            If True, will overwrite existing data array. Default is False.

        Returns
        -------
        Union[xarray.DataArray, xarray.Dataset]
            Data array or dataset with correct coordinate ordering.

        Examples
        --------
        >>> import xarray
        >>> import numpy
        >>> da = xarray.DataArray(
        ...  numpy.arange(16).reshape(4,4),
        ...  coords={"lat":numpy.array([87, 88, 89, 90]),
        ...          "lon":numpy.array([70, 69, 68, 67])}
        ... )
        >>> da_inv = da.aat.invert_coordinates()
        >>> da_inv.get_index("lon")
        Int64Index([67, 68, 69, 70], dtype='int64', name='lon')
        >>> da_inv.get_index("lat")
        Int64Index([90, 89, 88, 87], dtype='int64', name='lat')
        """
        data_obj = self._get_obj(inplace=inplace)
        lon_inv, lat_inv = self._check_coords_inverted()

        # Flip the raster as necessary (based on the flags)
        inv_dict = {}

        if lon_inv:
            logger.info("Longitude was inverted, reversing coordinates.")
            inv_dict[self.x_dim] = data_obj[self.x_dim][::-1]

        if lat_inv:
            logger.info("Latitude was inverted, reversing coordinates.")
            inv_dict[self.y_dim] = data_obj[self.y_dim][::-1]

        if inv_dict:
            data_obj = data_obj.reindex(inv_dict)

        return data_obj if not inplace else None

    def _check_coords_inverted(self):
        """
        Check if latitude and longitude inverted.

        Examples
        --------
        >>> import xarray
        >>> import numpy
        >>> da = xarray.DataArray(
        ...  numpy.arange(16).reshape(4,4),
        ...  coords={"lat":numpy.array([90, 89, 88, 87]),
        ...          "lon":numpy.array([70, 69, 68, 67])}
        ... )
        >>> da.aat._check_coords_inverted()
        (True, False)
        """
        data_obj = self._get_obj(inplace=False)
        lat = data_obj.get_index(self.y_dim)
        lon = data_obj.get_index(self.x_dim)
        return lon[0] > lon[-1], lat[0] < lat[-1]

    def change_longitude_range(
        self, to_180_range: bool = True, inplace: bool = False
    ) -> Optional[Union[xr.DataArray, xr.Dataset]]:
        """Convert longitude range between -180 to 180 and 0 to 360.

        The standard longitude range is from -180 to 180, while some
        applications use 0 to 360.

        ``change_longitude_range()`` will convert between the
        two coordinate ranges based on its current state.
        By default it will use the -180 to 180 range unless
        `to_180_range` is False, then it will use 0-360
        If coordinates lie solely between 0 and 180 then there is
        no need for conversion and the input  will be returned.

        Parameters
        ----------
        to_180_range : bool, default = True
            If True, the returned range is -180 to 180
            Else, the returned range is 0 to 360
        inplace : bool, optional
            If True, will overwrite existing data array. Default is False.

        Returns
        -------
        Union[xarray.DataArray, xarray.Dataset]
            Dataset with transformed longitude coordinates.

        Examples
        --------
        >>> import xarray
        >>> import numpy
        >>> temp = 15 + 8 * numpy.random.randn(4, 4, 3)
        >>> precip = 10 * np.random.rand(4, 4, 3)
        >>> ds = xarray.Dataset(
        ...   {
        ...     "temperature": (["lat", "lon", "time"], temp),
        ...     "precipitation": (["lat", "lon", "time"], precip)
        ...   },
        ...   coords={
        ...     "lat":numpy.array([87, 88, 89, 90]),
        ...     "lon":numpy.array([5, 120, 199, 360]),
        ...     "time": pd.date_range("2014-09-06", periods=3)
        ...   }
        ... )
        >>> ds_inv = ds.aat.change_longitude_range()
        >>> ds_inv.get_index("lon")
        Int64Index([-161, 0, 5, 120], dtype='int64', name='lon')
        >>> # invert coordinates back to original, in place
        >>> ds_inv.aat.change_longitude_range(inplace=True)
        >>> ds_inv.get_index("lon")
        Int64Index([0, 5, 120, 199], dtype='int64', name='lon')
        """
        data_obj = self._get_obj(inplace=inplace)

        if to_180_range and self.longitude_range == "360":
            logger.info("Converting longitude from 0 360 to -180 to 180.")

            data_obj[self.x_dim] = np.sort(
                ((data_obj[self.x_dim] + 180) % 360) - 180
            )
            self._longitude_range = "180"

        elif not to_180_range and self.longitude_range == "180":
            logger.info("Converting longitude from -180 to 180 to 0 to 360.")

            data_obj[self.x_dim] = np.sort(data_obj[self.x_dim] % 360)
            self._longitude_range = "360"

        else:
            logger.info("Coordinates already in required range.")

        return data_obj if not inplace else None


@xr.register_dataarray_accessor("aat")
class AatRasterArray(AatRasterMixin, RasterArray):
    """AA toolbox extension for xarray.DataArray."""

    def __init__(self, xarray_object):
        super().__init__(xarray_object)

    def compute_raster_stats(
        self,
        gdf: gpd.GeoDataFrame,
        feature_col: str,
        stats_list: Optional[List[str]] = None,
        percentile_list: Optional[List[int]] = None,
        all_touched: bool = False,
    ) -> pd.DataFrame:
        """Compute raster statistics for polygon geometry.

        ``compute_raster_stats()`` is designed to
        quickly compute raster statistics across a polygon
        and its features.

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            GeoDataFrame with row per area for stats computation.
            If ``pd.DataFrame`` is passed, geometry column must
            have the name ``geometry``.
        feature_col : str
            Column in ``gdf`` to use as row/feature identifier.
        stats_list : Optional[List[str]], optional
            List of statistics to calculate, by default None.
            Passed to ``get_attr()``.
        percentile_list : Optional[List[int]], optional
            List of percentiles to compute, by default None.
        all_touched : bool, optional
            If ``True`` all cells touching the region will be
            included, by default False. If ``False``, only
            cells with their centre in the region will be
            included.

        Returns
        -------
        pandas.DataFrame
            Dataframe with computed statistics.

        Examples
        --------
        >>> import geopandas as gpd
        >>> import xarray as xr
        >>> import rioxarray
        >>> from shapely.geometry import Polygon
        >>>
        >>> # compute raster stats on simple data
        >>> d = {
        ...     "name": ["area_a", "area_b"],
        ...     "geometry": [
        ...         Polygon([(0, 0), (0, 2), (2, 2), (2, 0)]),
        ...         Polygon([(2, 0), (2, 2), (3, 2), (3, 0)]),
        ...     ],
        ... }
        >>> gdf = gpd.GeoDataFrame(d)
        >>>
        >>> da = xr.DataArray(
        ...     [[1, 2, 3], [4, 5, 6]],
        ...     dims=("y", "x"),
        ...     coords={"y": [1.5, 0.5], "x": [0.5, 1.5, 2.5]},
        ... ).rio.write_crs("EPSG:4326")
        >>>
        >>> da.aat.compute_raster_stats(
        ...     gdf=gdf,
        ...     feature_col="name"
        ... ) # doctest: +SKIP
           mean_name            std_name min_name max_name sum_name count_name    name # noqa: E501
        0       3.0  1.5811388300841898        1        5     12.0          4  area_a  # noqa: E501
        1       4.5                 1.5        3        6      9.0          2  area_b  # noqa: E501
        """
        data_obj = self._get_obj(inplace=False)
        if data_obj.rio.crs is None:
            raise MissingCRS("No CRS found, set CRS before computation.")

        df_list = []

        if stats_list is None:
            stats_list = ["mean", "std", "min", "max", "sum", "count"]

        for feature in gdf[feature_col].unique():
            gdf_adm = gdf[gdf[feature_col] == feature]

            da_clip = data_obj.rio.set_spatial_dims(
                x_dim=self.x_dim, y_dim=self.y_dim
            )

            # clip returns error if no overlapping raster cells for geometry
            # so catching and skipping rest of iteration so no stats computed
            try:
                da_clip = da_clip.rio.clip(
                    gdf_adm.geometry, all_touched=all_touched
                )
            except NoDataInBounds:
                logger.warning(
                    "No overlapping raster cells for %s, skipping.", feature
                )
                continue

            grid_stat_all = []
            for stat in stats_list:
                # count automatically ignores NaNs
                # therefore skipna can also not be given as an argument
                # implemented count cause needed for computing percentages
                kwargs: Dict[str, Any] = {}
                if stat != "count":
                    kwargs["skipna"] = True
                # makes sum return NaN instead of 0 if array
                # only contains NaNs
                if stat == "sum":
                    kwargs["min_count"] = 1
                grid_stat = getattr(da_clip, stat)(
                    dim=[self.x_dim, self.y_dim], **kwargs
                ).rename(f"{stat}_{feature_col}")
                grid_stat_all.append(grid_stat)

            if percentile_list is not None:
                grid_quant = [
                    da_clip.quantile(quant / 100, dim=[self.x_dim, self.y_dim])
                    .drop("quantile")
                    .rename(f"{quant}quant_{feature_col}")
                    for quant in percentile_list
                ]
                grid_stat_all.extend(grid_quant)

            # if dims is 0, it throws an error when merging
            # and then converting to a df
            # this occurs when the input da is 2D
            if not grid_stat_all[0].dims:
                df_adm = pd.DataFrame(
                    {
                        da_stat.name: [da_stat.values]
                        for da_stat in grid_stat_all
                    }
                )

            else:
                zonal_stats_xr = xr.merge(grid_stat_all)
                df_adm = (
                    zonal_stats_xr.to_dataframe()
                    .drop("spatial_ref", axis=1)
                    .reset_index()
                )
            df_adm[feature_col] = feature
            df_list.append(df_adm)

        df_zonal_stats = pd.concat(df_list).reset_index(drop=True)
        return df_zonal_stats


@xr.register_dataset_accessor("aat")
class AatRasterDataset(AatRasterMixin, RasterDataset):
    """AA toolbox extension for xarray.Dataset."""

    def __init__(self, xarray_object):
        super().__init__(xarray_object)

    def get_raster_array(self, var_name: str) -> xr.DataArray:
        """Get xarray.DataArray from variable and keep dimensions.

        Accessing a component xarray.DataArray using the
        non-coordinate variable name loses and dimensions set
        through ``rio`` or ``aat``. This includes ``x_dim``,
        ``y_dim``, and ``t_dim`` that have to be specifically
        set using ``rio.set_spatial_dims()`` or
        ``aat.set_time_dim()`` respectively. For any dataset
        ``ds``, ``ds.get_raster_array("var")`` will retrieve
        the data array without losing the dimensions. Using
        ``ds["var"]`` will lose the dimensions.

        Parameters
        ----------
        var_name : str
            Name of variable.

        Returns
        -------
        xarray.DataArray
            A data array.

        Examples
        --------
        >>> import xarray
        >>> import numpy
        >>> temp = 15 + 8 * numpy.random.randn(4, 4, 3)
        >>> precip = 10 * np.random.rand(4, 4, 3)
        >>> ds = xarray.Dataset(
        ...   {
        ...     "temperature": (["lat", "lon", "F"], temp),
        ...     "precipitation": (["lat", "lon", "F"], precip)
        ...   },
        ...   coords={
        ...     "lat":numpy.array([87, 88, 89, 90]),
        ...     "lon":numpy.array([5, 120, 199, 360]),
        ...     "F": pd.date_range("2014-09-06", periods=3)
        ...   }
        ... )
        >>> ds.aat.set_time_dim("F", inplace=True)
        >>> da = ds.aat.get_raster_array("temperature")
        >>> da.aat.t_dim
        'F'
        >>> # directly accessing array loses set dimensions
        >>> ds['temperature'].aat.t_dim # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        rioxarray.exceptions.DimensionError: Time dimension not found.
            'aat.set_time_dim()' or using 'rename()' to change the
            dimension name to 't' can address this.
        Data variable: temperature
        """
        obj = self._obj[var_name]
        # rioxarray attributes
        if self._x_dim is not None and self._y_dim is not None:
            obj.rio.set_spatial_dims(
                x_dim=self._x_dim, y_dim=self._y_dim, inplace=True
            )

        if self.crs is not None:
            obj.rio.set_crs(self._crs, inplace=True)

        # raster module attributes
        if self._t_dim is not None:
            obj.aat.set_time_dim(self._t_dim, inplace=True)

        return obj

    def compute_raster_stats(
        self, var_names: Union[List[str], str, None] = None, **kwargs: Any
    ):
        """Compute raster statistics across dataset arrays.

        ``compute_raster_stats()`` calculates raster statistics on
        component data arrays of a dataset. By default, calculates on all
        non-coordinate variables, unless a list of variable names is
        passed in, which then have statistics calculated for them.

        Parameters
        ----------
        var_names : Union[List[str], str, None], optional
            Dataset data array variables to calculate raster statistics on.

        kwargs : Any
            Keyword arguments passed to the array method
            :meth:`~AatRasterArray.compute_raster_stats`

        Returns
        -------
        List[pandas.DataFrame]
            List of raster statistics data frames.

        Examples
        --------
        >>> import geopandas as gpd
        >>> import xarray as xr
        >>> import rioxarray
        >>> from shapely.geometry import Polygon
        >>>
        >>> # compute raster stats on simple data
        >>> d = {
        ...     "name": ["area_a", "area_b"],
        ...     "geometry": [
        ...         Polygon([(0, 0), (0, 2), (2, 2), (2, 0)]),
        ...         Polygon([(2, 0), (2, 2), (3, 2), (3, 0)]),
        ...     ],
        ... }
        >>> gdf = gpd.GeoDataFrame(d)
        >>>
        >>> ds = xr.DataArray(
        ...     [[1, 2, 3], [4, 5, 6]],
        ...     dims=("y", "x"),
        ...     coords={"y": [1.5, 0.5], "x": [0.5, 1.5, 2.5]},
        ... ).rio.write_crs("EPSG:4326").to_dataset(name="data")
        >>>
        >>> ds.aat.compute_raster_stats(
        ...    var_names=["data"],
        ...    gdf=gdf,
        ...    feature_col="name"
        ... ) # doctest: +SKIP
        [  mean_name            std_name min_name max_name sum_name count_name    name # noqa: E501
        0       3.0  1.5811388300841898        1        5     12.0          4  area_a  # noqa: E501
        1       4.5                 1.5        3        6      9.0          2  area_b] # noqa: E501
        """
        if var_names is None:
            var_names = self.vars
        elif isinstance(var_names, str):
            var_names = [var_names]

        stats = [
            self._obj[var].aat.compute_raster_stats(**kwargs)
            for var in var_names
        ]
        return stats
